#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Abstract classes definition module
#
"""
Abstract classes definition module.

Should not be needed by end users.
"""
import gtk

from events import EventHandler, Event

#===============================================================================
# Commands and command context
#===============================================================================
class Command(object):
    """
    Abstract class for commands.
    
    name must be a unique string identifier.
    """
    def __init__(self, name="generic command"):
        self.name = name
        self._commander = None
    
    def _check_commander(self, commander):
        """
        Check a commander.
        
        This method should raise a TypeError exception if commander if not
        appropriate for command.
        """
        pass
    
    def get_commander(self):
        """Return commander"""
        return self._commander
    
    def set_commander(self, commander, forced=False):
        """
        Set command's commander if it is appropriate for command. Otherwise,
        a TypeError is raised.
        
        if forced is set to True, no cheking is done
        """
        if not forced:
            self._check_commander(commander)
        self._commander = commander
    
    def setup(self):
        """
        Set command up.
        
        This method should be overridden if command needs to do something prior
        to be run.
        """
        pass
        
    def run(self):
        """
        Run command.
        
        Must be overridden.
        """
        raise NotImplementedError

class MenuItemCommand(Command):
    """
    Base class for Menu item command.
    
    name must be a unique command identifier string.
    item_entry must describe a menu item.
    menu_entry must describe the item parent's menu.
    item_group is string used as a gtk.ActionGroup's name for item.
      Optional with a default value of 'default'.
    menu_group is string used as a gtk.ActionGroup's name for menu.
      Optional with a default value of 'default'.
    
    Both item_entry and menu_entry must be a one-to-six sized tuple with the
    following information:
    
      * The name of the action. Must be specified.
    
      * The stock id for the action.
          Optional with a default value of None if a label is specified.
    
      * The label for the action.
          Optional with a default value of None if a stock id is specified.
    
      * The accelerator for the action, in the format understood by the
          gtk.accelerator_parse() function.
          Optional with a default value of None.
    
      * The tooltip for the action.
          Optional with a default value of None.
    
      * The callback function invoked when the action is activated.
          Optional with a default value of None.
    
    See: http://www.pygtk.org/docs/pygtk/class-gtkactiongroup.html
    """
    
    def __init__(self,
                 name,
                 item_entry,
                 menu_entry,
                 item_group="default",
                 menu_group="default",
                 item_position=None,
                 menu_position=None):
        Command.__init__(self, name)
        
        # Item
        self.item_entry = item_entry
        self.item_group = item_group
        self.item_position = item_position
        self.item_top_separator = False
        self.item_bottom_separator = False
        
        # Menu
        self.menu_entry = menu_entry
        self.menu_group = menu_group
        self.menu_position = menu_position
        self.menu_top_separator = False
        self.menu_bottom_separator = False
        
    def _check_commander(self, commander):
        """
        Check a commander.
        
        This method should raise a TypeError exception if commander if not
        appropriate for command.
        """
        if not isinstance(commander, BaseApplication):
            raise TypeError("%s  is not an application!" % repr(commander))
    
    def setup(self):
        """
        Set command up.
        
        Add menu and menu item to parent's application menu.
        """
        
        menu = self._commander.menu
        menu_path = self.menu_entry[0]
        
        separators = (self.menu_top_separator, self.menu_bottom_separator)
        menu.add_menu("",
                      self.menu_entry,
                      group=self.menu_group,
                      position=self.menu_position,
                      separators=separators
                      )
        
        separators = (self.item_top_separator, self.item_bottom_separator)
        menu.add_item("%s/" % menu_path,
                      self.item_entry,
                      group=self.item_group,
                      position=self.item_position,
                      separators=separators)
        

#===============================================================================
# Commander
#===============================================================================
class Commander(object):
    """
    Abstract class for commander interface.
    
    A commander maintains a set of commands and run them.
    """
    
    def __init__(self, commands=list(), hidden_commands=dict()):
        """
        Initialize Commander.
        """
        
        self._commands = list()
        self._commands_by_name = dict()
        
        self._hidden_commands = list()
        self._hidden_commands_by_name = dict()
        
        # Add commands
        for command in commands:
            self.add_command(command)
        
        for command in hidden_commands:
            self.add_hidden_command(command)
        
    
    def get_commands(self):
        """Return commands"""
        return list(self._commands)
    
    def add_command(self, command, setup=True, hidden=False):
        """
        Add command to commander.
        
        TODO: Complete docstring
        """
        name = command.name
        
        if hidden:
            commands = self._hidden_commands
            commands_by_name = self._hidden_commands_by_name
        else:
            commands = self._commands
            commands_by_name = self._commands_by_name
        
        if command.name in commands_by_name.keys():
            raise ValueError("Duplicated command name:%s" % name)
        
        if setup:
            # Updater command's commander
            command.set_commander(self)
            
            # Set command up
            command.setup()
        else:
            # Update command's commander
            command.set_commander(self, forced=True)
        
        # Finally add command
        commands.append(command)
        commands_by_name[name] = command
    
    def add_hidden_command(self, command, setup=True):
        """Add hidden command to commander"""
        self.add_command(command, setup=setup, hidden=True)
    
    def merge_from_commander(self, commander):
        """Add commands and context processors from another commander"""
        
        if not isinstance(commander, Commander):
            raise TypeError("Not a commander: %s" % repr(commander))
        
        for command in commander.get_commands():
            self.add_command(command)

    def run_command(self, name):
        """
        Run command using its name.
        """
        try:
            return self._commands_by_name[name].run()
        except KeyError:
            raise KeyError("No such command: %s" % name)

        


#===============================================================================
# Application component
#===============================================================================
class ApplicationComponent(EventHandler, Commander):
    """
    Application Component abstract class.
    
    TODO: complete docstring.
    """
    def __init__(self, name, widget=None, application=None):
        # Call ancestor __init__
        EventHandler.__init__(self)
        Commander.__init__(self)
        
        self.name = name
        
        # widget attribute should not be accessed directly
        self.widget = widget
        
        self.application = application
    
    def setup(self):
        """
        Run component setup.
        
        This method is implementation-dependent and must be overridden.
        """
        raise NotImplementedError
            
    
#===============================================================================
# Application events and command
#===============================================================================
class ExitEvent(Event):
    """Event triggered on exit"""
    name = "exit_event"

class ExitCommand(MenuItemCommand):
    """General Exit command"""
    def __init__(self):
        item_entry = ("exit",
                      gtk.STOCK_QUIT,
                      "E_xit",
                      None,
                      "Exit Program",
                      self.run
                      )
        item_group = "default"
        item_position = "bottom"
        
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "exit_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
        self.item_top_separator = True
    
    def run(self):
        """Run application's exit method"""
        return self._commander.exit()
    
#===============================================================================
# Application
#===============================================================================
class BaseApplication(EventHandler, Commander):
    """
    Application base class.
    
    TODO: complete docstring
    """
    def __init__(self, name="Generic application", version="1.0"):
        # Call ancestor __init__
        EventHandler.__init__(self)
        Commander.__init__(self)
        
        # Components handling
        self._components_by_name = dict()
        self._components = list()

        # Special components
        self.menu = None
        self.main_display = None
        self.status_bar = None
        
        # Application related
        self.name = name
        self.version = version
        
        # Gui related
        self.window = None
        
        # setup related
        self._is_setup = False
        self._is_gui_setup = False
    
    def _setup_window(self):
        """
        Window setup.
        
        Implementation-dependent. Must be overridden
        """
        raise NotImplementedError
    
    def _setup_commands(self):
        """Setup application commands"""
        commands = [ExitCommand()]
        
        for command in commands:
            self.add_command(command)
    
    def _setup_events(self):
        """Setup events"""
        self.add_event(ExitEvent)
          
    def setup(self, forced=False):
        """Application setup."""
        
        if not forced:
            if self._is_setup:
                return
        
        # Event setup
        self._setup_events()
        
        # Menu setup
        if self.menu is not None:
            self.menu.setup()
            self.merge_from_commander(self.menu)
            self.menu.application = self
        
        # Status bar setup
        if self.status_bar is not None:
            self.status_bar.setup()
            self.merge_from_commander(self.status_bar)
            self.status_bar.application = self
        
        # Main display setup
        if self.main_display is not None:
            self.main_display.setup()
            self.merge_from_commander(self.main_display)
            self.main_display.application = self
            
        # Setup core commands
        self._setup_commands()
        
        # Other components setup
        for component in self._components:
            component.setup()
            self.merge_from_commander(component)
            
#            for event in component.get_events():
#                self.add_event(event)
        self._is_setup = True
    
    def setup_gui(self):
        if self._is_gui_setup:
            return
        
        # GUI setup
        if self.window is not None:
            self._setup_window()
        
        self._is_gui_setup = True
        
        
    
    def run(self):
        """
        Run application.
        
        Only run setup() method. Should be extended.
        """
        if not self._is_setup:
            self.setup()
        
        if not self._is_gui_setup:
            self.setup_gui()
    
    def exit(self):
        """
        Exit application.
        
        Only propagate Exit event. Should be extended.
        """
        # Propagate ExitEvent
        event = ExitEvent()
        self._event_dispatcher(event)
        
    
    def get_component(self, name):
        """Get component by name"""
        try:
            return self._components_by_name[name]
        except KeyError:
            raise KeyError("No component registered with such name: %s" % name)
    
    def add_component(self, component):
        """Add a component to application"""
        if not isinstance(component, ApplicationComponent):
            raise TypeError("Not a valid component: %s" % repr(component))
        
        name = component.name
        
        if name in self._components_by_name:
            raise ValueError("'%s' component is already registered!" % name)
        
        self._components_by_name[name] = component
        self._components.append(component)
        
        component.application = self