#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# 
#
from __future__ import division

# global imports
import numpy
import os
import struct

# local imports
import spectrum
import util

def read_binary_file(filename, is_big_endian=False):
    """
    Read a binary file and extract a list containing number from file (format:long).
    Returns list if OK, empty list object if not.
    """
    try:
        fobj = open(filename, "rb")
        content = fobj.read()
        fobj.close()
    except IOError:
        return []
    
    data = list()

    for i in range(0, len(content), 4):
        if is_big_endian:
            data.append(struct.unpack(">l", content[i : i+4])[0])
        else:
            data.append(struct.unpack("<l", content[i : i+4])[0])
    return data

def read_paramaters(filepath, parameters=list()):
    """
    Read Bruker Parameter file and search for a list of parameter.
    Return a dictionnary with parameter names as keys and parameter values as
    values (values = None object if parameter).
    """
    try :
        result = dict()
        fobj = open(filepath, 'r')
    except IOError:
        pass
  
    else:
        for line in fobj:
            # Parameter and value extraction from line
            if line.find("##$") != -1: # this is a parameter line
                tmpparam = line[line.find("##$")+3 : line.find("=")]
                tmpvalue = line[line.find("= ")+2 :]
      
                if tmpparam in parameters:  # check if extracted parameter is useful
                    try:
                        result[tmpparam] = util.str_to_number(tmpvalue)
                    except ValueError:
                        result[tmpparam] = tmpvalue.strip()
        fobj.close()
        
    #complete result with unfound parameters
    for param in parameters:
        if not result.has_key(param):
            result[param] = None
        
    return result

def import_bruker_1r(pathname):
    """
    Imports a Bruker 1r file
    """
    
    # retrieve processing and acquisition folders
    pathname = os.path.realpath(pathname)
    proc_folder = os.path.dirname(pathname)
    acqu_folder = pathname[0 : pathname.rfind("pdata")]
    
    # 1st: Get acquisition parameters
    acqu_parameters = ["AQSEQ",
                       "BF1",
                       "BYTORDA",
                       "NS",
                       "NUC1",
                       "NC",
                       "O1",
                       "OVERFLW",
                       "SFO1",
                       "SW",
                       "SW_h",
                       "TD",
                       "TE"]
    acqu_parameters = read_paramaters(os.path.join(acqu_folder,"acqus"), acqu_parameters)
    
    # 2nd: Get processing parameters
    proc_parameters = ["BYTORDP",
                       "FTSIZE",
                       "LB",
                       "NC_proc",
                       "NSP",
                       "OFFSET",
                       "PHC0",
                       "PHC1",
                       "SI",
                       "SW_p",
                       "SF"]
    proc_parameters = read_paramaters(os.path.join(proc_folder, "procs"), proc_parameters)
    
    # 3rd: Read Real spectrum...
    re_spc = read_binary_file(os.path.join(proc_folder, "1r"), bool(proc_parameters["BYTORDP"]) )
    re_spc = numpy.array(re_spc)
    
    # 4th: Read Imaginary spectrum...
    im_spc = read_binary_file(os.path.join(proc_folder, "1i"), bool(proc_parameters["BYTORDP"]) )
    im_spc = numpy.array(im_spc)
    
    # 5th: Create complex spectrum
    spc = re_spc + im_spc * 1j
    
    # ... and correct spectrum intensity
    factor = 2 ** float(proc_parameters["NC_proc"])
    spc *= factor
    
    # 6th: Get title
    try :
        fobj = open(os.path.join(proc_folder, "title"), 'r')
    except IOError:
        title = u""
    else:
        title = fobj.read().strip()
        title = unicode(title,"iso8859-15")
        title = title.encode("utf-8")
        fobj.close()
    
    
    # Build axes
    llim_ppm = proc_parameters["OFFSET"]
    
    llim_hz = proc_parameters["OFFSET"] * acqu_parameters["BF1"]
    
    inc_hz =  - proc_parameters["SW_p"] / float(proc_parameters["SI"])
    inc_ppm =  inc_hz / acqu_parameters["BF1"]

    hz_axis = numpy.array([llim_hz + i * inc_hz for i in range(proc_parameters["SI"])])
    ppm_axis = numpy.array([llim_ppm + i * inc_ppm for i in range(proc_parameters["SI"])])
    
    # Estimate acquisition duration
    aq_time = 1 / (2 * acqu_parameters["SW_h"]) * acqu_parameters["TD"]
    
    tmp = os.path.dirname(acqu_folder).split(os.sep)
    
    name = "%s - exp #%s" %(tmp[-2], tmp[-1])
    
    # Finally build nmr spectrum
    nmr_spc = spectrum.NMRSpectrum(spc, ppm_axis, hz_axis, aq_time, title, name)
    
    return nmr_spc

if __name__ == "__main__":
    import sys
    import_bruker_1r(sys.argv[1])