#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Quick and Dirty Depaker main module
#
"""Quick and Dirty Depaker main module"""

from __future__ import division

# global import
import gtk
import os
import platform

# local imports
from constants import version, data_folder
import bruker_import
import mpl
from events import Event
from bases import ApplicationComponent, MenuItemCommand
import documentmanager
from application import Application
import spectrum
import util
import gtkutil

from fft_depake import fft_depake
from moments import calculate_moments

gpl_licence = \
"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

logo = gtk.gdk.pixbuf_new_from_file(os.path.join(data_folder,
                                                 "images",
                                                 "logo.png")
                                    )

#===============================================================================
# Specific events
#===============================================================================
class UpdateUIEvent(Event):
    name = "update_ui_event"
    def __init__(self, ui):
        # Call ancestor
        Event.__init__(self)
        self.ui = ui

#===============================================================================
# Fancy About dialog
#===============================================================================
class FancyAboutComponent(ApplicationComponent):
    """Class to handle display of information about an application"""
    def __init__(self, application=None):
        # Call ancestor __init__
        ApplicationComponent.__init__(self,
                                      name="fancy_about",
                                      widget=None,
                                      application=application)
        
    
    def setup(self):
        """Setup"""
        # Add command
        self.add_command(ShowFancyAboutCommand(), setup=False)
        
        # Set widget up
        self.widget = gtk.AboutDialog()
        self.widget.set_name(self.application.name)
        self.widget.set_version(self.application.version)
        self.widget.set_copyright(u"Copyright (c) 2009 Sébastien Buchoux")
        self.widget.set_website(u"https://launchpad.net/nmrdepaker")
        self.widget.set_comments("De-Pake-ing made easy!")
        self.widget.set_license(gpl_licence)
        self.widget.set_authors([u"Sébastien Buchoux"])
        self.widget.set_logo(logo)
        self.widget.set_artists([u"Anne-Sophie Bruyant",
                                 u"Sébastien Buchoux"])

class ShowFancyAboutCommand(MenuItemCommand):
    """General Exit command"""
    def __init__(self):
        
        # Item
        item_entry = ("about",
                      gtk.STOCK_ABOUT,
                      "A_bout...",
                      None,
                      "About this program",
                      self.run)
        item_group = "default"
        item_position = "bottom"
        
        # Menu
        menu_entry = ("Help", None, "_Help")
        menu_group = "default"
        menu_position = "bottom"
        
        MenuItemCommand.__init__(self,
                                 "fancy_about_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
    def run(self):
        """Show about component's widget"""
        about = self._commander.get_component("fancy_about")
        
        # Run GTK Dialog
        about.widget.run()
        
        # Hide dialog
        about.widget.hide()

#===================================================================================================
# Depaker specific document handling commands
#===================================================================================================
class DepakerOpenCommand(documentmanager.OpenCommand):
    def __init__(self):
        documentmanager.OpenCommand.__init__(self)
        
        self.file_filters = {"Bruker 1r file": "1r"}
        
    def _opener(self, pathname):
        self._commander.document_manager.load_document(pathname)

class DepakerSaveCommand(documentmanager.SaveCommand):        
    def run(self):
        """Show about component's widget"""
        application = self._commander
        document = application.document_manager.active_document
        file_filters = {"Text file" : "*.txt"}
        
        pathname = gtkutil.path_from_chooser(title="Save as...",
                                             action="save",
                                             file_filters=file_filters)
        
        if pathname is not None:
            extension = os.path.splitext(pathname)[1]
            correct_extension = file_filters.values()[0][1:]
            if extension != correct_extension:
                pathname += correct_extension
            
            document.pathname = pathname
            application.document_manager.save_document(document)

class DepakerExportCommand(documentmanager.ExportCommand):
    def _filter_from_document(self, document):
        # Retrieve canvas
        canvas = self._get_canvas()
        
        file_filters = {} 
        
        if canvas is not None:
            for format, description in canvas.get_supported_filetypes().items():
                file_filters["%s (%s)"% (description, format)] = "*.%s" % format
        
        self.file_filters = file_filters
    
    def _get_canvas(self):
        notebook = self._commander.document_manager.widget
        view_widget = notebook.get_nth_page(notebook.get_current_page())
        
        for child in view_widget.get_children():
            if isinstance(child, mpl.nmrfigure.NMRCanvas):
                return child
        
        return None
    
    def _exporter(self, document, pathname):
        format = os.path.splitext(pathname)[1][1:]
        canvas = self._get_canvas()
        try:
            canvas.print_figure(pathname, format=format)
        except Exception:
            gtkutil.message_die("Error occurred while exporting window", "error")
        
#===================================================================================================
# Depaker specific document view
#===================================================================================================
class DepakerSpectrumView(documentmanager.DocumentView):
    def __init__(self, document, document_manager):
        documentmanager.DocumentView.__init__(self, document, document_manager)
       
        self.toolbar = None
        self.canvas = None
        
        self.spectrum_id = -1
        self.moments = None
    
    def _setup_view(self):
        self.view = gtk.VBox()
        
        # Spectrum view setup
        self.canvas = mpl.nmrfigure.NMRCanvas()
        self.toolbar = mpl.nmrfigure.NMRCanvasToolbar(self.canvas,
                                                      application=self,
                                                      pix_folder=os.path.join(data_folder,
                                                                              "images")
                                                      )

        self.view.pack_start(self.toolbar, False, False)
        self.view.pack_start(self.canvas, True, True)
    
    def update_view(self):
        spectrum = self._document.get_spectrum()
        
        if self.spectrum_id == -1:
            self.spectrum_id = self.canvas.add_plot(spectrum.x,
                                                    spectrum.y,
                                                    "b",
                                                    label="spectrum")
        else:
            self.canvas.update_plot_data(self.spectrum_id,
                                         spectrum.x,
                                         spectrum.y)
        
        # Add legend to figure
        self.canvas.draw_idle()
    
    def _check_limits(self, llim, rlim, pivot):
        x_axis = self._document.get_spectrum().x
        
        left = util.closest_index(x_axis, llim)
        right = util.closest_index(x_axis, rlim)
        
        pivot = util.closest_index(x_axis, pivot)
        
        if left > right:
            left, right = right, left
        
        if pivot < left or pivot > right:
            gtkutil.message_die("Pivot must be within calculation limits!", "warning")
            return (-1, -1, -1)
        
        if right - left < 100:
            gtkutil.message_die("Left and right calculation limits too close!", "warning")
            return (-1, -1, -1)
        
        return (left, right, pivot)
    
    def depake(self):
        nmr_spc = self._document.get_spectrum()
        spc = nmr_spc.spectrum
        hz_axis = nmr_spc.hz_axis
        ppm_axis = nmr_spc.ppm_axis
        aq_time = nmr_spc.aq_time
        
        hz_axis, ppm_axis, depaked_spc = fft_depake(hz_axis, ppm_axis, spc, aq_time)
        
        # Build new document
        nmr_spc = spectrum.NMRSpectrum(depaked_spc, ppm_axis, hz_axis, aq_time, title="Depaked spectrum")
        document = spectrum.NMRSpectrumDocument(nmr_spc, name = "%s (Depaked)" % self._document.name )
        
        # Load new document
        self._document_manager.load_document(document)
        
        
    def moment_calculation(self, llim, rlim, pivot):
        left, right, pivot = self._check_limits(llim, rlim, pivot)
        
        if left == -1:
            return
        
        spc = self._document.get_spectrum().y
        hz_axis = self._document.get_spectrum().x
        
        moments = calculate_moments(hz_axis, spc, left, right, pivot)
        
        results = "Limits used (Hz): %.0f to %.0f\n" % (moments["left_limit"],
                                                        moments["right_limit"]
                                                        )
        results += "Pivot: %.0f Hz\n" % moments["pivot"]
        results += "M1 zero: %.0f Hz\n\n" % moments["M1_zero"]
        
        results += "M1 using pivot: %.0f Hz\n" % moments["M1_with_pivot"]
        
        results += "M1 using M1 zero: %.0f Hz\n\n" % moments["M1_with_zero"]
        
        results += "M1sym using pivot: %.3e Hz\n" % moments["M1sym_with_pivot"]
        
        results += "M1sym using M1 zero: %.3e Hz\n\n" % moments["M1sym_with_zero"]
        
        results += "M2 using pivot: %.3e Hz²\n" % moments["M2_with_pivot"]
        
        results += "M2 using M1 zero: %.3e Hz²\n\n" % moments["M2_with_zero"]
        
        # Display results
        dialog = gtk.MessageDialog(parent=None,
                               type=gtk.MESSAGE_INFO,
                               buttons=gtk.BUTTONS_OK,
                               message_format="")
        
        dialog.set_title("Moment calculation results")
        dialog.set_markup(results)
        dialog.set_border_width(10)
        dialog.set_transient_for(None)
        
        dialog.set_position(gtk.WIN_POS_CENTER)
        dialog.set_gravity(gtk.gdk.GRAVITY_NORTH)
        
        dialog.show()
        
        dialog.connect("delete-event", lambda *args: dialog.destroy())
        dialog.connect("response", lambda *args: dialog.destroy())
        
        
        

#===================================================================================================
# Depaker specific document manager
#===================================================================================================
class DepakerDocumentManager(documentmanager.DocumentManagerComponent):
    def __init__(self):
        documentmanager.DocumentManagerComponent.__init__(self)
        
        # Add Spectrum View
        self.view_types["NMR spectrum"] = DepakerSpectrumView
        
        # Override document command
        self.document_command_classes["import"] = None
        self.document_command_classes["open"] = DepakerOpenCommand
        self.document_command_classes["save"] = DepakerSaveCommand
        self.document_command_classes["save_as"] = None
        self.document_command_classes["export"] = DepakerExportCommand
    
    def _document_loader(self, document_path):
        if isinstance(document_path, basestring):
            nmr_spc = bruker_import.import_bruker_1r(document_path)
            
            document = spectrum.NMRSpectrumDocument(nmr_spc, nmr_spc.name)
        elif isinstance(document_path, spectrum.NMRSpectrumDocument):
            document = document_path
        else:
            raise TypeError("Not a QaD depaker document: %s" % document_path)
        
        return document
        
    def _document_saver(self, document):
        
        nmr_spc =  document.get_spectrum()
        
        content = u"#Name: %s\n#Description: %s\n" % (document.name,
                                                       nmr_spc.title)
        
        content += u"#Hz\t\tPPM\t\tIntensity\n"
        
        hz_axis = nmr_spc.hz_axis
        ppm_axis = nmr_spc.ppm_axis
        y = nmr_spc.y

        for i, val in enumerate(y):
            content += "%.3f\t%.3f\t%.3f\n" % (hz_axis[i], ppm_axis[i], val)
        
        
        fname = document.pathname
        fobj = open(fname, "w")
        fobj.write(content)
        fobj.close()
    
#===================================================================================================
# Depaker main application
#===================================================================================================
class DepakerApplication(Application):
    def __init__(self, version=version):
        Application.__init__(self, "NMR Depaker", version)
        self.main_display = DepakerDocumentManager()
        self.document_manager = self.main_display
        self.add_component(FancyAboutComponent())
    
    def _setup_window(self):
        Application._setup_window(self)
        
        self.window.set_icon_from_file(os.path.join(data_folder, "images", "icon.png"))

    def _setup_events(self):
        # Call ancestor's method
        Application._setup_events(self)
        
        self.add_event(UpdateUIEvent)
        
    def setup(self, forced=False):
        # Call ancestor setup
        Application.setup(self, forced)
        
        # Connect events
        self.document_manager.connect("switch_document_event", self.update_ui)
        
        # Update UI
        self.update_ui()
    
    def update_ui(self, event=None):
        # ensure menu entries related to active document are insensitive
        if self.document_manager.active_document is None:
            self.menu.set_group_sensitive("active_document", False)
        else:
            self.menu.set_group_sensitive("active_document", True)
        
        # Dispatch event
        event = UpdateUIEvent(self.menu)
        self._event_dispatcher(event)
    
    def get_active_document(self):
        try:
            return self.document_manager.active_document
        except:
            return None
    
    def exit(self):
        modified = False
        for document in self.document_manager.documents:
            if document.is_modified():
                modified = True
                break
        if modified:
            message = "Exit without saving modifications?"
            dialog = gtk.MessageDialog(parent=None,
                                       flags=gtk.DIALOG_MODAL,
                                       type=gtk.MESSAGE_QUESTION,
                                       buttons=gtk.BUTTONS_YES_NO,
                                       message_format=message)
            dialog.set_border_width(10)
    
            dialog.set_position(gtk.WIN_POS_CENTER)
            dialog.set_gravity(gtk.gdk.GRAVITY_NORTH)

            
            resp = dialog.run()
            dialog.destroy()
            if resp == gtk.RESPONSE_NO:
                return True # Stop event propagation
        
        Application.exit(self)
        return False

if __name__ == "__main__":
    app = DepakerApplication()
    app.run()
