#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Document manager module
#
"""Document manager module"""
import gobject
import gtk
from os import path
from time import localtime

from bases import ApplicationComponent, MenuItemCommand
from events import Event, EventHandler
import gtkutil
import util

#===============================================================================
# Events
#===============================================================================
class SwitchDocumentEvent(Event):
    """Event triggered when active document is switch"""
    name = "switch_document_event"
    def __init__(self, document):
        # Call ancestor
        Event.__init__(self)
        self.document = document

class DocumentLoadedEvent(Event):
    """Event triggered when document is loaded"""
    name = "document_loaded_event"
    def __init__(self, document):
        # Call ancestor
        Event.__init__(self)
        self.document = document

class DocumentUnLoadingEvent(Event):
    """Event triggered before document is unloaded"""
    name = "document_unloading_event"
    def __init__(self, document):
        # Call ancestor
        Event.__init__(self)
        self.document = document

class DocumentChangedEvent(Event):
    """Event triggered when active document is changed"""
    name = "document_changed_event"
    def __init__(self, document):
        # Call ancestor
        Event.__init__(self)
        self.document = document

#===============================================================================
# Commands
#===============================================================================
class OpenCommand(MenuItemCommand):
    """General Open command"""
    def __init__(self):
        
        # Item
        item_entry = ("open",
                      gtk.STOCK_OPEN,
                      "_Open...",
                      None,
                      "Open a file",
                      self.run)
        item_group = "default"
        item_position = None
        
        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "open_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
        self.file_filters = {"All files": "*.*"}
        
    def run(self):
        """Show about component's widget"""
        pathname = gtkutil.path_from_chooser(title="Select a file to open",
                                             action="open",
                                             file_filters=self.file_filters)
        
        if pathname is not None:
            self._opener(pathname)
    
    def _opener(self, pathname):
        raise NotImplementedError

class ImportCommand(MenuItemCommand):
    """General import command"""
    def __init__(self):
        
        # Item
        item_entry = ("import",
                      gtk.STOCK_CONVERT,
                      "_Import...",
                      None,
                      "Import a file",
                      self.run)
        item_group = "default"
        item_position = None
        
        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "import_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
        self.file_filters = {"All files": "*.*"}
        
    def run(self):
        """Show about component's widget"""
        pathname = gtkutil.path_from_chooser(title="Select a file to import",
                                             action="open",
                                             file_filters=self.file_filters)
        
        if pathname is not None:
            self._importer(pathname)
    
    def _importer(self, pathname):
        raise NotImplementedError
            
class SaveCommand(MenuItemCommand):
    """General save command"""
    def __init__(self):
        
        # Item
        item_entry = ("save",
                      gtk.STOCK_SAVE,
                      "_Save",
                      None,
                      "Save the current file",
                      self.run)
        item_group = "active_document"
        item_position = None

        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "save_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
    def run(self):
        """Show about component's widget"""
        application = self._commander
        document = application.document_manager.active_document
        
        if document.pathname == "":
            self._commander.run_command("save_as_command")
        else:
            application.document_manager.save_document(document)

class SaveAsCommand(MenuItemCommand):
    """General 'save as' command"""
    def __init__(self):
        
        # Item
        item_entry = ("save_as",
                      gtk.STOCK_SAVE_AS,
                      "Save _As...",
                      None,
                      "Save the current file with a different name",
                      self.run)
        item_group = "active_document"
        item_position = None

        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "save_as_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
        self.file_filters = {"All files": "*.*"}
        
    def run(self):
        """Show about component's widget"""
        application = self._commander
        document = application.document_manager.active_document
        self._filter_from_document(document)
        
        pathname = gtkutil.path_from_chooser(title="Save as...",
                                             action="save",
                                             file_filters=self.file_filters)
        
        if pathname is not None:
            extension = path.splitext(pathname)[1]
            correct_extension = self.file_filters.values()[0][1:]
            if extension != correct_extension:
                pathname += correct_extension
            
            document.pathname = pathname
            application.document_manager.save_document(document)
    
    def _filter_from_document(self, document):
        pass

class ExportCommand(MenuItemCommand):
    """General 'export as' command"""
    def __init__(self):
        
        # Item
        item_entry = ("export",
                      None,
                      "_Export...",
                      None,
                      "Export the current file",
                      self.run)
        item_group = "active_document"
        item_position = None

        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "export_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
        self.file_filters = {"All files": "*.*"}
        
    def run(self):
        """Show about component's widget"""
        application = self._commander
        document = application.document_manager.active_document
        self._filter_from_document(document)
        
        
        pathname = gtkutil.path_from_chooser(title="Export as...",
                                             action="save",
                                             file_filters=self.file_filters)
        
        if pathname is not None:
            self._exporter(document, pathname)
            
    def _filter_from_document(self, document):
        pass
    
    def _exporter(self, document, pathname):
        raise NotImplementedError

class CloseCommand(MenuItemCommand):
    """General close command"""
    def __init__(self):
        
        # Item
        item_entry = ("close",
                      gtk.STOCK_CLOSE,
                      "_Close",
                      None,
                      "Close the current file",
                      self.run)
        item_group = "active_document"
        item_position = None

        # Menu
        menu_entry = ("File", None, "_File")
        menu_group = "default"
        menu_position = "top"
        
        MenuItemCommand.__init__(self,
                                 "close_command",
                                 item_entry,
                                 menu_entry,
                                 item_group=item_group,
                                 item_position=item_position,
                                 menu_group=menu_group,
                                 menu_position=menu_position,
                                 )
        
    def run(self):
        """Show about component's widget"""
        self._commander.document_manager.unload_document()

#===============================================================================
# Document types
#===============================================================================
class Document(EventHandler):
    def __init__(self, name="undefined", document_type="generic"):
        events = [DocumentChangedEvent]
        EventHandler.__init__(self, events)
        
        # RW variables
        self.name = name
        
        self.pathname = ""
        
        # RO or special variables
        self._modified = False
        self._type = document_type
        
    def _dispatch_modification(self):
        # Dispatch event
        event = DocumentChangedEvent(self)
        self._event_dispatcher(event)
    
    def is_modified(self, modified = None):
        if modified is None:
            return self._modified
        else:
            self._modified = bool(modified)
            
        if modified is True:
            self._dispatch_modification()
    
    modified = property(is_modified, is_modified)
    
    def get_type(self):
        return self._type
    
    type = property(get_type)
     
#===============================================================================
# Document views
#===============================================================================
class DocumentView(EventHandler):
    """Handle a generic document view"""
    def __init__(self, document, document_manager):
        # Call ancestor __init__
        EventHandler.__init__(self)
        
        self.view = None
        self.label = None
        self._document = document
        self._document_manager = document_manager
        
        if document is not None:
            self.cid = document.connect("document_changed_event", self.update_cb)
        else:
            self.cid = None
    
    def _setup_label(self):
        """Set document label up"""
        self.label = gtk.Label(self._document.name)
    
    def _setup_view(self):
        """Set document view up"""
        self.view = gtk.VBox()
        label = gtk.Label()
        self.view.pack_start(label, True, True)
        
    def setup(self):
        """
        Set component up.
        
        Should not be overridden.
        """
        self._setup_label()
        self._setup_view()
        
        self.update_label()
        self.update_view()
        
    def get_document(self):
        """Return document"""
        return self._document
    
    def set_document(self, document):
        """Set current document"""
        # Disconnect from old document
        if self.cid is not None:
            self._document.disconnect(self.cid)
            self.cid = None
        
        self._document = document
        
        if document is not None:
            self.cid = document.connect("document_changed_event", self.update_cb)
            
        self.update()
    
    def update_cb(self, *args):
        print "Updating View"
        self.update()
    
    def update(self):
        self.update_label()
        self.update_view()
    
    def update_label(self):
        """
        Update document label.
        """
        if self.label is None:
            self._setup_label()
        
        if self._document.is_modified():
            name = "*%s" % self._document.name
        else:
            name = self._document.name
        
        self.label.set_text(name)
        #print "New label:", name, self._document._modified
        
    def update_view(self):
        """
        Update document view.
        
        Should be overridden."""
        
        if self.view is None:
            self._setup_view()
        
        label = self.view.get_children()[0]
        
        label.set_text("Generic view of document: %s" % repr(self._document))
        

#===============================================================================
# Application components    
#===============================================================================
class DocumentManagerComponent(ApplicationComponent):
    """Component to handle documents"""
    def __init__(self, application=None):
        # Call ancestor __init__
        ApplicationComponent.__init__(self,
                                      name="document_manager",
                                      widget=None,
                                      application=application)
        
        self.view_types = {"generic" : DocumentView}
        self.documents = list()
        self.document_views = list()
        
        self.active_document = None
        
        self.document_command_classes = {"open": OpenCommand,
                                         "close": CloseCommand,
                                         "import": ImportCommand,
                                         "export": ExportCommand,
                                         "save": SaveCommand,
                                         "save_as": SaveAsCommand,
                                         }
        
    def load_document(self, document_path, set_active=True):
        """Load a document"""
        
        # Load document if necessary
        document = self._document_loader(document_path)
        
        if document in self.documents:
            raise ValueError("Document already loaded: %s" % repr(document))
        
        try:
            doc_type = document.get_type()
        except AttributeError:
            doc_type = "generic"
        
        try:
            view = self.view_types[doc_type](document, self)
            view.setup()
            
            # show gtk widgets
            view.view.show_all()
            view.label.show_all()
            
        except KeyError:
            gtkutil.message_die("Unhandled document type: %s" % doc_type,
                                "warning")
            return
        
        self.documents.append(document)
        self.document_views.append(view)
        
        # Update notebook
        page_tab = gtk.HBox(False, 2)
        
        # build a button with a close icon
        image = gtk.image_new_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        tab_button = gtk.Button()
        tab_button.add(image)
        tab_button.set_relief(gtk.RELIEF_NONE)
        
        tab_button.connect("clicked", self.unload_document_cb, document)
        page_tab.pack_start(view.label, False)
        page_tab.pack_start(tab_button, False)
        
        page_tab.show_all()
        
        if set_active:
            page_id = self.widget.append_page(view.view, page_tab)
        else:
            page_id = self.widget.get_current_page()
            self.widget.append_page(view.view, page_tab)
            
        self.widget.set_current_page(page_id)
        
        # dispatch loaded event
        event = DocumentLoadedEvent(document)
        self._event_dispatcher(event)
    
    def _document_loader(self, document_path):
        raise NotImplementedError
    
    def save_document(self, document):
        """Save document"""
        self._document_saver(document)
        
        index = self.documents.index(document)
        view = self.document_views[index]
        view.update()
        
        # Dispatch document changed event
        event = DocumentChangedEvent(document)
        self._event_dispatcher(event)
    
    def _document_saver(self, document):
        raise NotImplementedError
          
    def setup(self):
        """About setup"""
        # Add events
        self.add_event(SwitchDocumentEvent)
        self.add_event(DocumentLoadedEvent)
        self.add_event(DocumentUnLoadingEvent)
        self.add_event(DocumentChangedEvent)
        
        # Add commands
        try:
            open_command = self.document_command_classes["open"]()
        except TypeError:
            pass
        else:
            open_command.item_top_separator = True
            self.add_command(open_command, setup=False)
            
        try:   
            self.add_command(self.document_command_classes["import"](), setup=False)
        except TypeError:
            pass
        
        try:
            close_command = self.document_command_classes["close"]()
        except TypeError:
            pass
        else:
            close_command.item_top_separator = True
            self.add_command(close_command, setup=False)
        
        try:
            save_command = self.document_command_classes["save"]()
        except TypeError:
            pass
        else:
            save_command.item_top_separator = True
            self.add_command(save_command, setup=False)
        
        try:
            self.add_command(self.document_command_classes["save_as"](), setup=False)
        except TypeError:
            pass
        
        try:
            self.add_command(self.document_command_classes["export"](), setup=False)
        except TypeError:
            pass
        
        # Set widget up
        self.widget = gtk.Notebook()
        self.widget.connect("switch-page", self.set_active_document_cb)
    
    def set_active_document_cb(self, notebook, not_used, index):
        """GTK-related callback to update active document"""
        
        page = notebook.get_nth_page(index)
        
        active_document = None
        for view in self.document_views:
            if page == view.view:
                active_document = view.get_document()
                break
            
        if active_document is None:
            raise ValueError("No such document view: %s" % repr(page))
        
        self.set_active_document(active_document)    
    
    def set_active_document(self, document):
        """Set active document"""
        
        self.active_document = document
        
        # Dispatch document switched event
        event = SwitchDocumentEvent(document)
        self._event_dispatcher(event)
        #print "New active document:", document
    
    def unload_document_cb(self, widget, document):
        self.unload_document(document)
    
    def unload_document(self, document=None):
        if document is None:
            document = self.active_document
        
        if document is None:
            return
        
        if document.is_modified():
            # Raise a popup to ask if document needs to be saved
            message = "Save %s before closing it?" % document.get_name()
            dialog = gtk.MessageDialog(parent=None,
                                       flags=gtk.DIALOG_MODAL,
                                       type=gtk.MESSAGE_QUESTION,
                                       buttons=gtk.BUTTONS_YES_NO,
                                       message_format=message)
            dialog.set_border_width(10)
    
            dialog.set_position(gtk.WIN_POS_CENTER)
            dialog.set_gravity(gtk.gdk.GRAVITY_NORTH)
            
            dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

            
            resp = dialog.run()
            dialog.destroy()
            if resp == gtk.RESPONSE_CANCEL:
                return
            elif resp == gtk.RESPONSE_YES:
                self.run_command("save_command")

        
        # Dispatch document unloading event
        event = DocumentUnLoadingEvent(document)
        self._event_dispatcher(event)
        
        index = self.documents.index(document)
        
        # Delete document reference
        del self.documents[index]
        view = self.document_views.pop(index)
        
        # Remove corresponding notebook page
        page_num = self.widget.page_num(view.view)
        self.widget.remove_page(page_num)
        
        # Delete view reference
        del view
        
        # Check if last page was removed (i.e. switch-page signal not triggered)
        if self.widget.get_n_pages() == 0:
            self.set_active_document(None)
            
    def update_document_view(self, document):
        if document.is_modified():
            # Dispatch document changed event
            event = DocumentChangedEvent(document)

    
