#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Event handling module
#
"""
Event handling module
"""

class Event(object):
    """Abstract class for events"""
    name = "generic_event"
    def __init__(self):
        self.name = self.__class__.name
    
    def __repr__(self):
        return "< Event instance: '%s' at %s >" % (self.name, str(id(self)))
    
    def __str__(self):
        return "'%s' Event" % self.name


# Following class is highly adapted from matplotib.cbook
class CallbackRegistry:
    """
    Handle registering and disconnecting for a set of signals.
    """
    def __init__(self, signals=list()):
        """
        *signals* is a sequence of signals.
        """
        self._cid = 0
        self.callbacks = dict()
        self.signals = list()
        
        for signal in signals:
            self.add_signal(signal)
    
    def add_signal(self, signal):
        """
        Add *signal* to accepted signals.
        
        ValueError is raised if signal is already registered
        """
        if type(signal) != str:
            raise TypeError("signal must be a string")
        if signal in self.callbacks.keys():
            raise ValueError("Signal is already registered: %s" % signal)
        else:
            self.callbacks[signal] = dict()
            self.signals.append(signal)
    
    def _check_signal(self, signal):
        """Make sure *signal* is a valid signal or raise a ValueError"""
        if signal not in self.signals:
            text = "'%s' signal is not valid. Valid signals: %s" % (signal,
                                                                   self.signals)
            raise ValueError(text)

    def connect(self, signal, func):
        """
        Register *func* to be called when a signal *signal* is generated.
        *func* will be called with an event as a single argument.
        
        Return a callback id.
        """
        self._check_signal(signal)
        self._cid += 1
        self.callbacks[signal][self._cid] = func
        return self._cid

    def disconnect(self, callback_id):
        """
        Disconnect the callback registered with *callback_id*
        """
        for callbackd in self.callbacks.values():
            try:
                del callbackd[callback_id]
            except KeyError:
                continue
            else:
                break

    def process(self, event):
        """
        Process signal related to *event*.
        
        Functions registered to receive callbacks on signal will be called.
        Functions must accept *event* as single argument.
        """
        signal = event.name
        self._check_signal(signal)
        for func in self.callbacks[signal].values():
            #print "Calling connected function:", func
            func(event)


class EventHandler(object):
    """Base class for event handlers"""
    
    def __init__(self, events=list()):
        self._callbacks = CallbackRegistry()
        self._events = list()
        
        for event in events:
            self.add_event(event)
    
    def connect(self, signal, func):
        """
        Register *func* to be called when a signal *signal* is generated.
        *func* will be called with an event as single argument.
        
        Return a callback id.
        """
        return self._callbacks.connect(signal, func)
    
    def disconnect(self, callback_id):
        """
        Disconnect the callback registered with *callback_id*
        """
        return self._callbacks.disconnect(callback_id)
    
    def add_event(self, event):
        """Add an event to event handler"""
        klass = event
        if not hasattr(klass, "mro"): # trigger is an instance, not a class
            klass = event.__class__
        if issubclass(klass, Event):
            self._callbacks.add_signal(klass.name)
            self._events.append(event)
        else:
            raise TypeError("Not a valid event: %s" % repr(event))
        
    def signals(self):
        """
        Return a list of valid signals.
        """
        return list(self._callbacks.signals)
    
    def get_events(self):
        """
        Return a list of valid events.
        """
        return list(self._events)
    
    def _event_dispatcher(self, event):
        """
        Process *event*
        """
        #print "%s triggered by %s" % (str(event), repr(self))
        self._callbacks.process(event)
