#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Miscellaneous utility functions
#
import os
import gtk
import pango
import traceback
import sys
from cStringIO import StringIO


def message_die(message="Nothing to say!", message_type="info"):
    """Launch a gtk.MessageDialog to display message.
    If message type is "warning" or "error", a traceback lookup may be done to gather information.
    
    Keyword arguments:
    message -- message to display (default "Nothing to say!")
    message_type -- one of the following: "info", "warning", "error" (default "info")
    
    """
    
    if message_type == "info":
        message_type = gtk.MESSAGE_INFO
        title = "Information"
    elif message_type == "warning":
        message_type = gtk.MESSAGE_WARNING
        title = "Warning"
        message = "<b>" + message + "</b>"
    else :
        message_type = gtk.MESSAGE_ERROR
        title = "Error!"
        message = "<b>" + message + "</b>"
    
    # Create dialog
    dialog = gtk.MessageDialog(parent=None,
                               flags=gtk.DIALOG_MODAL,
                               type=message_type,
                               buttons=gtk.BUTTONS_NONE,
                               message_format="")
    dialog.set_title(title)
    dialog.set_markup(message)
    dialog.set_border_width(10)
    
    dialog.set_position(gtk.WIN_POS_CENTER)
    dialog.set_gravity(gtk.gdk.GRAVITY_NORTH)
    
    dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
    
    dialog.run()
    dialog.destroy()

def fancy_excepthook(type, value, tb):
    if type == KeyboardInterrupt:
        sys.exit("KeyboardInterrupt")
    
    elif type == NotImplementedError:
        dialog = gtk.MessageDialog(parent=None,
                                   flags=gtk.DIALOG_MODAL,
                                   type=gtk.MESSAGE_INFO,
                                   buttons=gtk.BUTTONS_CLOSE,
                                   message_format="")
        dialog.set_title("Feature not implemented yet!")
        if len(str(value)) > 0:
            dialog.set_markup(str(value))
        else:
            dialog.set_markup("Sorry, this feature is not\nimplemented yet!")
        dialog.set_size_request(300, 150)
        dialog.run()
        dialog.destroy()

    else:
        string = "<b><big>An unhandled exception has been detected hooked</big></b>.\n\nIt is not fatal but it may affect current process."
        dialog = gtk.MessageDialog(parent=None,
                                   flags=gtk.DIALOG_MODAL,
                                   type=gtk.MESSAGE_WARNING,
                                   buttons=gtk.BUTTONS_NONE,
                                   message_format="")
        dialog.set_title("Unexpected error!")
        dialog.set_markup(string)
        dialog.set_resizable(True)
        dialog.set_size_request(400, 150)
        
        show_button = dialog.add_button("Show Details...", 1)
        hide_button = dialog.add_button("Hide Details", 2)
        hide_button.hide()
        
        dialog.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
        
        # Details
        textview = gtk.TextView()
        textview.show()
        textview.set_editable(False)
        textview.modify_font(pango.FontDescription("Monospace"))
        
        sw = gtk.ScrolledWindow()
        sw.show()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add(textview)
        
        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_IN)
        frame.add(sw)
        frame.set_border_width(6)
        
        dialog.vbox.add(frame)
        
        textbuffer = textview.get_buffer()
        trace = StringIO()
        traceback.print_exception(type, value, tb, None, trace)

        textbuffer.set_text(trace.getvalue())
        textview.set_size_request(width=-1, height=150)
        
        dialog.details = frame
        dialog.set_position(gtk.WIN_POS_CENTER)
        dialog.set_gravity(gtk.gdk.GRAVITY_NORTH)
        
        while 1:
            resp = dialog.run()
            if resp == 1:
                dialog.set_size_request(400, 350)
                dialog.details.show()
                show_button.hide()
                hide_button.show()
            elif resp == 2:
                dialog.set_size_request(400, 150)
                dialog.details.hide()
                show_button.show()
                hide_button.hide()
            else: break
        dialog.destroy()
        sys.exit(-1)

def path_from_chooser(title="Please select location", action="open", file_filters={"All files": ["*"]}, start_path="~"):

    """
    Open a window with a file chooser et return the file (or folder) path.
    title is a string containing dialog title
    action is one of the following string: 'select_folder', 'save' or 'open'
    filefilter is a list containing file extension desired
    """
    
    # Get correct action for dialog and set correct buttons to use
    if action == "folder":
      gtkaction = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER
      gtkbuttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK)
      
    elif action == "save":
      gtkaction = gtk.FILE_CHOOSER_ACTION_SAVE
      gtkbuttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,gtk.STOCK_SAVE, gtk.RESPONSE_OK)
      
    else:
      gtkaction=gtk.FILE_CHOOSER_ACTION_OPEN
      gtkbuttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK)

    # Build appropriate file filter list
    filterlist=[]
    
    for name, patterns in file_filters.items():
        filter = gtk.FileFilter()
        filter.set_name(name)
        
        if type(patterns) == list:
            for pattern in patterns:
                filter.add_pattern(pattern)
        else:
            filter.add_pattern(patterns)
        
        filterlist.append(filter)
    
    # create dialog
    dialog = gtk.FileChooserDialog(title,
                               None,
                               gtkaction,
                               gtkbuttons)
    dialog.set_default_response(gtk.RESPONSE_OK)
    dialog.set_do_overwrite_confirmation(True)
    
    #dialog.set_size_request(300, 400)
    
    # add filter to filechooser
    for filter in filterlist:
      dialog.add_filter(filter)
    
    # select correct folder
    start_path = os.path.realpath(os.path.expanduser(start_path))
    dialog.set_current_folder(start_path)
    
    # set current name
    #dialog.set_current_name ('')
    
    # show dialog
    response = dialog.run()
    
    # get selected file (if selected!)
    selected_file = None
    if response == gtk.RESPONSE_OK:
      selected_file = unicode(dialog.get_filename())
      normal_extension = file_filters[dialog.get_filter().get_name()][1:]
      
      if os.path.splitext(selected_file)[1] != normal_extension:
          selected_file += normal_extension
    else:
      pass

    dialog.destroy()
    
    
    return selected_file
