#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009  Sébastien Buchoux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# NMR spectrum display machinery
#
"""NMR spectrum display machinery"""
from __future__ import division
import gtk
from math import cos, sin, pi
from matplotlib.figure import Figure
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg
from matplotlib.lines import Line2D
from matplotlib.legend import Legend
from matplotlib.text import Text
from matplotlib.patches import Rectangle, Polygon
#from matplotlib.artist import Artist
import matplotlib.transforms as mpl_transforms
from numpy import array
import os


#===============================================================================
# Util functions
#===============================================================================
def _triangle_verts_from_xy(x,y, size=3, orientation=0):
    theta = pi * 2.0/3.0 # 120 degree
    orientation = (orientation / 180.0) * pi 
    verts = []
    
    for i in range(3):
        verts.append([x + size * cos(orientation + theta * i),
                      y + size * sin(orientation + theta * i)])
    
    # copy first point to close path
    verts.append(verts[0])
    
    return array(verts)

def _zoom_factor(value, direction="in", factor=0.1):
    sgn = value / abs(value)
    factor = 1 + factor * sgn
    
    if direction == "in":
        return 1.0 / factor
    elif direction == "out":
        return factor
    else:
        raise ValueError("Not a valid zoom direction: %s", str(direction))
    
#===============================================================================
# NMR canvas class
#===============================================================================
class NMRCanvas(FigureCanvasGTKAgg):
    """GTK Drawing area to render NMR plots"""
    
    def __init__(self):
        # Create matplotlib Figure (empty graph)
        self.figure = Figure(facecolor="w")
        
        # Execute Ancestors class init
        FigureCanvasGTKAgg.__init__(self, self.figure) # a gtk.DrawingArea
        
        # Initialize axes
        self.axes = self.figure.add_subplot(111)
        
        # Initialize axes labels
        self.x_label = None
        self.y_label = None
        
        # Internal variables
        self._plots = {}
        self._movable_boxes = {}
        self._legend = None
        self._center = None
        
        # Internal variables related to user interaction
        self._modifier = None
        self._modifier_press = None
        
        self._mode = None
        self._press = None
        self._grid_on = True
        self._pan_direction = None
        
        self.working = False
        
        # Internal variables related to object dragging/interaction
        self._draggable = None
        self._active = None
        self._dragged_press = None
        self._dragged_bbox = None
        
        # Internal variables related to animation
        self._background = None
        
        # Setup figure
        self.setup_axes()
        
        # Connect matplotlib events
        self.mpl_connect('draw_event', self.mpl_draw_cb)
        self.mpl_connect('button_press_event', self.mpl_button_press_cb)
        self.mpl_connect('key_press_event', self.mpl_key_press_cb)
        self.mpl_connect('key_release_event', self.mpl_key_release_cb)
        self.mpl_connect('scroll_event', self.mpl_scroll_event_cb)
        
        # matplotlib callback ids
        self._motion_cid = None
        self._release_cid = None
        self._picker_cid = None
        
        self._restore_default_cb()

#===============================================================================
# Util methods
#===============================================================================
    def _find_draggable(self, event):
        """Return draggable object or None"""
        
        # Check if some objects are under pointer
        draggable = dict()
        for mpl_object in self._movable_boxes.values():
            if isinstance(mpl_object, Legend): # little hack for legend
                under, details = mpl_object.legendPatch.contains(event)
            else:
                under, details = mpl_object.contains(event)
            
            if under == True:
                draggable[mpl_object.get_zorder()] = mpl_object
            
        # Select draggable object
        if len(draggable) > 0:
            draggable = draggable[max(draggable.keys())]
        else:
            draggable = None
        
        return draggable
    
    def _store_background(self):
        self._background = self.copy_from_bbox(self.axes.bbox)

    def _update_animated_artists(self, artists):
        if type(artists) not in (list, tuple):
            artists = [artists,]
            
        # Be sure all artists are set to animated state
        force_redraw = False
        for artist in artists:
            if not artist.get_animated():
                artist.set_animated(True)
                force_redraw = True
        
        if force_redraw: 
            # Force redraw now
            self.draw()

        self.restore_region(self._background)
        
        for artist in artists:
            self.axes.draw_artist(artist)
        
        self.blit(self.axes.bbox)
        
    def _restore_default_cb(self):
        # Update callbacks
        if self._motion_cid is not None:
            self.mpl_disconnect(self._motion_cid)
        self._motion_cid = self.mpl_connect("motion_notify_event",
                                            self.mpl_generic_motion_cb)
        
        if self._release_cid is not None:
            self.mpl_disconnect(self._release_cid)
        self._release_cid = None
        
        if self._picker_cid is not None:
            self.mpl_disconnect(self._picker_cid)
        self._picker_cid = None
    
    def cancel_interaction(self, interaction):
        if interaction == "x-box-zoom_confirm":
            self.mpl_span_picker_cb(None)
        elif interaction == "box-zoom_confirm":
            self.mpl_zoom_picker_cb(None)
        else:
            print "WARNING: Should cancel", interaction, "mode"
        
    def reset_mode(self):        
        # Force redraw
        self.draw_idle()
        
        # Restore callbacks
        self._restore_default_cb()
        
        # Unset mode
        self._mode = None
              
    def set_mode(self, mode):
        self._mode = mode
        
        return True
    
#===============================================================================
# Figure setup methods
#===============================================================================
    def setup_axes(self):
        """Axes setup"""
        # Initialize unit labels
        x_label = u"Hz"
        self.x_label = self.axes.set_xlabel(x_label)
        y_label = u"Intensity (A.U)"
        self.y_label = self.axes.set_ylabel(y_label)
        for text in self.axes.get_ymajorticklabels():
            text.set_visible(False)
        self.axes.grid(b=self._grid_on, color="#99BBFF", antialiased=True)
        
        self.figure.subplots_adjust(right=0.95, top=0.95)

    def zoom(self, axis="Y", direction="in" ):
        
        xmin, xmax = self.axes.get_xlim()
        ymin, ymax = self.axes.get_ylim()
        
        if "X" in axis:
            xmin *= _zoom_factor(xmin, direction, 0.05)
            xmax /= _zoom_factor(xmax, direction, 0.05)
            
        if "Y" in axis:
            ymin /= _zoom_factor(ymin, direction)
            ymax *= _zoom_factor(ymax, direction)
        
        # Update axes limits
        self.axes.set_xlim(xmin, xmax)
        self.axes.set_ylim(ymin, ymax)
        
        # Force redraw
        self.draw_idle()
    
    def zoom_on_data(self):
        xmin = xmax = ymin = ymax = None
        for i, plot in enumerate(self._plots.values()):
            x_data = plot.get_xdata()
            y_data = plot.get_ydata()
            
            if i == 0:
                xmin = min(x_data)
                xmax = max(x_data)
                
                ymin = min(y_data)
                ymax = max(y_data)
            
            if min(x_data) < xmin:
                xmin = min(x_data)
            
            if max(x_data) > xmax:
                xmax = max(x_data)
            
            if min(y_data) < ymin:
                ymin = min(y_data)
            
            if max(y_data) > ymax:
                ymax = max(y_data)
        
        self.axes.set_xlim(xmax, xmin)
        self.axes.set_ylim(ymin*1.1, ymax*1.1)
        
        # Force redraw
        self.draw_idle()
    
    def set_movable(self, mpl_object):
        id = len(self._movable_boxes) + 1
        if isinstance(mpl_object, (Legend, Text)):
            
            if not isinstance(mpl_object, Legend):
                if mpl_object.get_transform() != self.axes.transAxes:
                    print "WARNING: Transform may not be compatible!"
                    mpl_object.set_transform(self.axes.transAxes)
                    
            self._movable_boxes[id] = mpl_object
        elif isinstance(mpl_object, Line2D):
            self._movable_boxes[id] = mpl_object
        else:
            raise TypeError("Not a matplotlib artist:", mpl_object)
    
    def toggle_grid(self, value=None):
        if value is None:
            self._grid_on = not self._grid_on
        else:
            self._grid_on = value
        
        # Update grid state
        self.axes.grid(b=self._grid_on)
        
        # Force redraw
        self.draw_idle()
        
#===============================================================================
# Content handling methods
#===============================================================================
    def update_plot_data(self, plot_id, x, y):
        """Update plot"""
        x = array(x)
        y = array(y)
        
        line = self._plots[plot_id]
        
        line.set_xdata(x)
        line.set_ydata(y)
        
        self.zoom_on_data()

    def get_legend(self):
        return self._legend
    
    def show_legend(self):
        """Add legend to figure"""
        if self._legend is None:
            self._legend = self.axes.legend(loc=4)
            
            self.set_movable(self._legend)
            
        self.set_movable(self._legend)
        return self._legend
    
    def update_legend(self):
        if self._legend is None:
            self.show_legend()
        
        lines = []
        labels = []
        
        plot_ids = self._plots.keys()
        plot_ids.sort()
        for plot_id in plot_ids:
            line = self._plots[plot_id]
            
            if line.get_visible():
                lines.append(line)
                labels.append(line.get_label())
        
        if len(lines) > 0:
            old_visible = self._legend.get_visible()
            
            # Update legend
            self._legend = self.axes.legend(lines, labels, loc=4)
            self.set_movable(self._legend)
            
            self._legend.set_visible(old_visible)
        else:
            self._legend.set_visible(False)
        
        self.draw_idle()
    
    def add_plot(self, x, y, *args, **kwargs):
        """Add plot to axes"""
        
        id = len(self._plots) + 1
        lines = self.axes.plot(x,y, *args, **kwargs)
        self._plots[id] = lines[0]
        
        self.zoom_on_data()
        return id
    
    def get_plot(self, plot_id):
        try:
            return self._plots[plot_id]
        except KeyError:
            return None
    
    def add_text(self, text, x=0.5, y=0.5):
        if len(text) == 0:
            return
        #x, y = self.axes.transAxes.transform([x,y])
        mpl_text = self.axes.text(x, y,
                                  text,
                                  ha="center",
                                  va="center",
                                  ma="right",
                                  transform=self.axes.transAxes,
                                  #transform=mpl_transforms.IdentityTransform(),
                                  bbox=dict(boxstyle="round",
                                            fc="b",
                                            ec="k",
                                            alpha=0.1,
                                            snap=True),
                                  
                                  )
        
        self.set_movable(mpl_text)
        return mpl_text
    
    def add_limits(self):
        
        
        # Retrieve X positions in data coordinates
        transdata = self.axes.transData
        transaxes = self.axes.transAxes
        
        l_x, y = transaxes.transform([0.25, 0])
        r_x, y = transaxes.transform([0.75, 0])
        
        l_x, y = transdata.inverted().transform([l_x, 0])
        r_x, y = transdata.inverted().transform([r_x, 0])
        
        # Add left limit
        lline = self.axes.axvline(l_x,
                                  color="#99EE99",
                                  linewidth=2)
        self.set_movable(lline)
        
        # Add left limit
        rline = self.axes.axvline(r_x,
                                  color="#99EE99",
                                  linewidth=2)
        self.set_movable(rline)
        
        # Add pivot
        pivot = self.axes.axvline((l_x + r_x) / 2,
                                  ymin=0.9,
                                  ymax=1,
                                  color="#FF5555",
                                  linewidth=1.5)
        self.set_movable(pivot)
        
        # Force redraw
        self.draw_idle()
        
        return (lline, rline, pivot)

#===============================================================================
# Matplotlib callbacks   
#===============================================================================
    # Callback called when draw event is triggered
    def mpl_draw_cb(self, event):
        """Callback called when draw event is triggered"""
        if self.working:
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
            
        # be sure focus is acquired
        if not self.is_focus():
            self.grab_focus()
            
        bboxes = []
        
        # Handle y axis title
        y_title_bbox = self.y_label.get_window_extent()
        bboxes.append(y_title_bbox.inverse_transformed(self.figure.transFigure))
        
        # Handle y ticks labels
        for label in self.axes.get_yticklabels():
            bbox = label.get_window_extent()
            # the figure transform goes from relative coords->pixels and we
            # want the inverse of that
            bboxi = bbox.inverse_transformed(self.figure.transFigure)
       
            bboxes.append(bboxi)
            
        # this is the bbox that bounds all the bboxes, again in relative
        # figure coords
        bbox = mpl_transforms.Bbox.union(bboxes)
        if self.figure.subplotpars.left < bbox.width:
            # we need to move it over
            self.figure.subplots_adjust(left=1.3*bbox.width) # pad a little
            self.figure.canvas.draw()
        elif self.figure.subplotpars.left > bbox.width*1.3:
            self.figure.subplots_adjust(left=1.3*bbox.width) # pad a little
            self.figure.canvas.draw()
        
        # Update background for animations
        self._store_background()
        
        # Update center position
        self._center = self.axes.transAxes.transform([0.5,0.5])
        
        return False
    
    # Key-press related callbacks
    def mpl_key_press_cb(self, event):
        """Callback called when a key is pressed"""
        if self.working:
            return
        
        key = event.key
        
        if key in ["control", "shift"]:
            self._modifier = key
            self._modifier_press = event.x, event.y
        elif key == "g":
            # Toggle grid
            self.toggle_grid()
        elif key == "escape":
            if self._mode in ["x-box-zoom_confirm", "box-zoom_confirm"]:
                self.cancel_interaction(self._mode)
        else:
            print key, "was pressed"
        
    def mpl_key_release_cb(self, event):
        """Callback called when a key is released"""
        if self.working:
            return
        self._modifier = None
        self._modifier_press = None
    
    # Callbacks called when a mouse button is pressed
    def mpl_scroll_event_cb(self, event):
        """
        Callback called when mouse wheel is used.
        Actually, this callback only calls the mpl_button_press_cb callback.
        """
        self.mpl_button_press_cb(event)
    
    def mpl_button_press_cb(self, event):
        """Generic button press event handler"""
        
        if event.inaxes==None: return # No interaction if click is not in axes
        if self.working:
            return
        
        # Store event coordinates
        self._press = event.x, event.y
        
        # New callbacks
        new_motion_cb = None
        new_release_cb = None
        
        # New interaction mode
        mode = self._mode
        
        # Handle special interaction modes that may be canceled in which case
        # normal mouse interaction must be supported 
        if mode == "x-box-zoom_confirm":
            if self.mpl_span_picker_cb(event):
                mode = "interaction-done"
            else: # zoom not confirmed
                mode = None
        elif mode == "box-zoom_confirm":
            if self.mpl_zoom_picker_cb(event):
                mode = "interaction-done"
            else: # zoom not confirmed
                mode = None
        else:
            if mode is not None:
                print "Set interaction according to mode:", mode
        
        # Set interaction mode according to mouse button pressed
        if mode is None:
            if event.button == 1: # left click
                if self._modifier == "control":
                    mode = "box-zoom"
                elif self._modifier == "shift":
                    mode = "x-box-zoom"
                else:
                    if self._draggable is not None:
                        mode = "drag"
                    else:
                        mode = "measure"
            elif event.button == 2: # middle click
                mode = "pan"
            elif event.button == "up": # mouse wheel up
                if self._modifier == "control":
                    mode = "zoom-in"
                elif self._modifier == "shift":
                    mode = "x-zoom-in"   
                else:
                    mode = "y-zoom-in"
            elif event.button == "down": # mouse wheel down
                if self._modifier == "control":
                    mode = "zoom-out"
                elif self._modifier == "shift":
                    mode = "x-zoom-out"   
                else:
                    mode = "y-zoom-out"

        # Act according to interaction mode
        if mode == "drag":            
            # Start dragging object
            self._active = self._draggable
            
            # Update cursor
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))
            
            # Get dragged object coordinates
            if isinstance(self._active, Text):
                x0, y0 = self._active.get_position()
            elif isinstance(self._active, Legend):
                patch = self._active.legendPatch
                x0 = patch.get_x()
                y0 = patch.get_y()
            elif isinstance(self._active, Line2D):
                x0 = self._active.get_xdata()[0]
                y0 = self._active.get_ydata()[0]
            else:
                raise TypeError("Can't get object position")
            
            if not isinstance(self._active, Line2D):
                # Convert to pixel
                x0, y0 = self._active.get_transform().transform([x0, y0])
            
                # Store dragged object coordinates
                self._dragged_press = x0, y0
                
                # Store dragged object bbox (pixel)
                bbox = self._active.get_window_extent()
                self._dragged_bbox = bbox.xmin, bbox.xmax, bbox.ymin, bbox.ymax
                
                # Animate artist
                #self._animate_artists(self._active)
            else:
                pass #Quick and dirty hack
                
            # update new callbacks
            new_motion_cb = self.mpl_drag_motion_cb
            new_release_cb = self.mpl_drag_release_cb
            
        elif mode == "x-box-zoom":
            # Build transform for span selector
            trans = mpl_transforms.blended_transform_factory(self.axes.transData,
                                                             self.axes.transAxes)
            
            # Create span selector
            self._active = Rectangle((0,0), # position
                                     0, # width
                                     1, # height
                                     transform=trans,
                                     visible=True,
                                     facecolor='blue',
                                     alpha=0.1
                                     )
            
            # update new callbacks
            new_motion_cb = self.mpl_span_motion_cb
            new_release_cb = self.mpl_span_release_cb
        
        elif mode == "box-zoom":            
            # Create span selector
            self._active = Rectangle((0,0), # position
                                     0, # width
                                     1, # height
                                     transform=self.axes.transData,
                                     visible=True,
                                     facecolor='blue',
                                     alpha=0.1
                                     )
            
            # update new callbacks
            new_motion_cb = self.mpl_zoom_motion_cb
            new_release_cb = self.mpl_zoom_release_cb
        
        elif mode == "pan":
            self.axes.start_pan(event.x, event.y, 1)
            
            # Update cursor
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))
            
            # update new callbacks
            new_motion_cb = self.mpl_pan_motion_cb
            new_release_cb = self.mpl_pan_release_cb
        
        elif mode == "measure":
            # Build transform
            #trans = mpl_transforms.blended_transform_factory(mpl_transforms.IdentityTransform(),
                                                             #self.axes.transAxes)
            
            color = "#aaaaff"
            alpha = 1.0
            lline = self.axes.axvline(event.xdata,
                                      color=color,
                                      visible=False,)
            
            rline = self.axes.axvline(event.xdata,
                                      color=color,
                                      visible=False,)
            
            arrow_body = Line2D([event.x, event.x],
                                [self._center[1], self._center[1]],
                                linewidth=1.0,
                                color=color,
                                alpha=alpha,
                                transform=mpl_transforms.IdentityTransform(),
                                )
            self.axes.add_artist(arrow_body)
            
            arrow_lend = Polygon(_triangle_verts_from_xy(event.x-3,
                                                         self._center[1],
                                                         3,
                                                         0),
                                   closed=True,
                                   alpha=alpha,
                                   fc=color,
                                   ec=color,
                                   visible=False,
                                   transform=mpl_transforms.IdentityTransform()
                                  )
            self.axes.add_artist(arrow_lend)
            
            arrow_rend = Polygon(_triangle_verts_from_xy(event.x+3,
                                                         self._center[1],
                                                         3,
                                                         180),
                                   closed=True,
                                   alpha=alpha,
                                   fc=color,
                                   ec=color,
                                   visible=False,
                                   transform=mpl_transforms.IdentityTransform()
                                  )
            
            self.axes.add_artist(arrow_rend)
            
            
            text = self.axes.text(event.x,
                                  self._center[1] + 10,
                                  "%.2f nm" % 0.0,
                                  visible=False,
                                  transform=mpl_transforms.IdentityTransform(),
                                  bbox=dict(boxstyle="round",
                                            fc="w",
                                            ec="w",
                                            alpha=0.5)
                                  )
            text.set_x(event.x - text.get_window_extent(renderer=self.get_renderer()).width /2.0)
            
            self._active = [lline,
                            rline,
                            arrow_body,
                            arrow_lend,
                            arrow_rend,
                            text]
            
            # Animate artists
            #self._animate_artists(self._active)
            
            # update new callbacks
            new_motion_cb = self.mpl_measure_motion_cb
            new_release_cb = self.mpl_measure_release_cb
            
        elif mode == "y-zoom-out":
            self.zoom(axis="Y", direction="out")
        elif mode == "y-zoom-in":
            self.zoom(axis="Y", direction="in")
        elif mode == "x-zoom-out":
            self.zoom(axis="X", direction="out")
        elif mode == "x-zoom-in":
            self.zoom(axis="X", direction="in")
        elif mode == "zoom-out":
            self.zoom(axis="XY", direction="out")
        elif mode == "zoom-in":
            self.zoom(axis="XY", direction="in")
        elif mode == "interaction-done":
            mode = None
        elif mode is None:
            pass
        else:
            print "WARNING: %s interaction not handled!" % mode
        
        # Update callbacks
        if new_motion_cb is not None:
            if self._motion_cid is not None:
                self.mpl_disconnect(self._motion_cid)
            self._motion_cid = self.mpl_connect("motion_notify_event",
                                                new_motion_cb)
        
        if new_release_cb is not None:
            if self._release_cid is not None:
                self.mpl_disconnect(self._release_cid)
            self._release_cid = self.mpl_connect("button_release_event",
                                                 new_release_cb)
        
        # Update canvas
        self.draw_idle()

    # Generic motion notify callback
    def mpl_generic_motion_cb(self, event):
        """
        Callback called when no specific motion_notify_event handler are
        specified.
        """
        
        if event.inaxes==None: # No interaction if motion is not in axes
            return
        if self.working:
            return
        
        # Default cursor type
        cursor = gtk.gdk.ARROW
        
        if self._mode in [None, "x-box-zoom_confirm"]:
            # Check if mouse is over a draggable object
            self._draggable = self._find_draggable(event)
            
            if self._draggable is not None:
                cursor = gtk.gdk.HAND2
        
        if self._mode in ["x-box-zoom_confirm", "box-zoom_confirm"]:
            if len(self._active.hitlist(event)) > 0: # use hitlist method because contains method always return True
                cursor = gtk.gdk.HAND2
        
        if self._mode in ["pan"]:
            cursor = gtk.gdk.HAND2
        
        # Update cursor
        self.window.set_cursor(gtk.gdk.Cursor(cursor))

            
    
    # Motion notify callback used when an object is dragged
    def mpl_drag_motion_cb(self, event):
        """
        motion_notify_event's callback called when an object is dragged.
        """
        if self.working:
            return
        
        if isinstance(self._active, Line2D):
            x_data = event.xdata # data_coord
            y_data = event.ydata
            
            self._active.set_xdata([x_data, x_data])
        else:
            # Retrieve button press coordinates
            x_press, y_press = self._press # pixel
        
            # Get current position
            x_motion, y_motion = event.x, event.y # pixel
        
            # Calculate deltas
            dx = x_motion - x_press # pixel
            dy = y_motion - y_press # pixel
            
            # Retrieve object bbox limits
            bbox_xmin, bbox_xmax, bbox_ymin, bbox_ymax = self._dragged_bbox
            
            # Retrieve axes limits
            min_x, min_y = self.axes.transAxes.transform([0.01, 0.01])
            max_x, max_y = self.axes.transAxes.transform([0.99, 0.99])
            
            # Snap to grid trick
            grid_size = int(min((max_x - min_x)/50, (max_y - min_y)/50 ))
            if grid_size > 1:
                dx = int (round(dx / grid_size) * grid_size)
                dy = int (round(dy / grid_size) * grid_size)
            
            # Correct dx and dy to stay within figure
            if bbox_xmin + dx < min_x:
                dx = min_x - bbox_xmin
                
            elif bbox_xmax + dx > max_x:
                dx = max_x - bbox_xmax
            
            if bbox_ymin + dy < min_y:
                dy = min_y - bbox_ymin
            
            elif bbox_ymax + dy > max_y:
                dy = max_y - bbox_ymax
                
            # Calculate new object coordinates 
            x0, y0 = self._dragged_press
            new_x, new_y = x0 + dx, y0 + dy
            
            # Retrieve the proper transformer
            transform = self.axes.transAxes.inverted().transform
            
            # Convert pixel to axes coords
            new_x, new_y = transform([new_x, new_y])
            
            # Update object position
            if isinstance(self._active, Text):
                self._active.set_x(new_x)
                self._active.set_y(new_y)
            elif isinstance(self._active, Legend):
                self._active._loc = (new_x, new_y)
            
            
        self._update_animated_artists(self._active)
        
    
    # Callback called when mouse button is released when an object is dragged
    def mpl_drag_release_cb(self, event):
        """
        Callback that handle button mouse release when an object is dragged.
        """
        if self.working:
            return
        
        # Update cursor
        if event.inaxes == None: # No longer over object
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
        else: # Still over object
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
        
        self._active.set_animated(False)
        self._active = None
        self._dragged_bbox = None
        self._dragged_press = None
        self._press = None
        self._background = None
        
        # Unset mode
        self.reset_mode()
    
    # Motion notify callback used when span selector in on
    def mpl_span_motion_cb(self, event):
        if event.inaxes == None:
            return
        
        if self.working:
            return
        
        # Retrieve data coords from stored press event coords
        x_press, y_press = self.axes.transData.inverted().transform(self._press)
        
        pos =  min(x_press, event.xdata)
        width = max(x_press, event.xdata) - pos
        
        self._active.set_x(pos)
        self._active.set_width(width)
        
        self._update_animated_artists(self._active)
        
    def mpl_span_release_cb(self, event):
        if self.working:
            return
        
        # Add span to axes
        self.axes.add_artist(self._active)
        
        # No more animations
        self._active.set_animated(False)
        
        # Force redraw
        self.draw_idle()
        
        # Restore callbacks
        self._restore_default_cb()
        
        # Unset mode
        self.set_mode("x-box-zoom_confirm")
    
    def mpl_span_picker_cb(self, event=None):
        if self.working:
            return
        
        interaction_confirmed = False
        if event is None:
            hits = list()
        else:
            hits =self._active.hitlist(event)
            
        if len(hits) > 0: # use hitlist method because contains method always return True
            interaction_confirmed = True
            
            # Retrieve rectangle's bbox
            bbox = self._active.get_bbox()
            
            # Set limits
            self.axes.set_xlim(bbox.xmax, bbox.xmin)
            
            # Reset cursor
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
        
        # Disconnect picker and restore default callbacks
        self._restore_default_cb()
        
        # Delete span selector if registered
        try:
            self._active.remove()
        except ValueError:
            pass
        
        # Unset mode
        self.reset_mode()
        
        return interaction_confirmed
    
    # Motion notify callback used when zoom selector in on
    def mpl_zoom_motion_cb(self, event):
        if event.inaxes == None:
            return
        if self.working:
            return
        
        # Retrieve data coords from stored press event coords
        x_press, y_press = self.axes.transData.inverted().transform(self._press)
        
        x_pos =  min(x_press, event.xdata)
        y_pos =  min(y_press, event.ydata)
        
        width = abs(x_press - event.xdata)
        height = abs(y_press - event.ydata)
        
        self._active.set_x(x_pos)
        self._active.set_y(y_pos)
        self._active.set_width(width)
        self._active.set_height(height)
        
        self._update_animated_artists(self._active)
        
    def mpl_zoom_release_cb(self, event):
        if self.working:
            return
        # Add span to axes
        self.axes.add_artist(self._active)
        
        # No more animations
        self._active.set_animated(False)
        
        # Force redraw
        self.draw_idle()
        
        # Restore callbacks
        self._restore_default_cb()
        
        # Unset mode
        self.set_mode("box-zoom_confirm")
    
    def mpl_zoom_picker_cb(self, event=None):
        if self.working:
            return
        
        interaction_confirmed = False
        if event is None:
            hits = list()
        else:
            hits = self._active.hitlist(event)
            
        if len(hits) > 0: # use hitlist method because contains method always return True
            interaction_confirmed = True
            
            # Retrieve rectangle's bbox
            bbox = self._active.get_bbox()
            
            # Set limits
            self.axes.set_xlim(bbox.xmax, bbox.xmin)
            self.axes.set_ylim(bbox.ymin, bbox.ymax)
            
            # Reset cursor
            self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
        
        # Disconnect picker and restore default callbacks
        self._restore_default_cb()
        
        # Delete span selector if registered
        try:
            self._active.remove()
        except ValueError:
            pass
        
        # Unset mode
        self.reset_mode()
        
        return interaction_confirmed
    
    # Motion notify callback used when pan in on
    def mpl_pan_motion_cb(self, event):
        if self.working:
            return
        
        if self._modifier_press is None:
            x_press = self._press[0]
            y_press = self._press[1]
        else:
            x_press = self._modifier_press[0]
            y_press = self._modifier_press[1]
            
        if self._modifier == "shift":
            if self._pan_direction is None:
                    
                dx = abs(event.x - x_press)
                dy = abs(event.y - y_press)
                
                if dx < dy:
                    self._pan_direction = "vert"
                else:
                    self._pan_direction = "horz"
            
            if self._pan_direction == "vert":
                x = x_press
                y = event.y
            else:
                x = event.x
                y = y_press
        else:
            self._pan_direction = None
            x = event.x
            y = event.y
            
        self.axes.drag_pan(1, None, x, y)
        
        self.draw_idle()
    
    def mpl_pan_release_cb(self, event):
        if self.working:
            return
        # Disconnect picker and restore default callbacks
        self._restore_default_cb()
        
        # Unset mode
        self.reset_mode()
        
        # Update cursor
        self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
    
    def mpl_measure_motion_cb(self, event):
        if event.inaxes == None:
            return
        if self.working:
            return
        
        
        # Retrieve mpl objects
        lline, rline, arrow_body, arrow_lend, arrow_rend, text = self._active
        
        # Check visibility
        if not lline.get_visible():
            lline.set_visible(True)
            rline.set_visible(True)
            text.set_visible(True)
            arrow_body.set_visible(True)
        
        # Retrieve coords from stored press event coords
        x_press, y_press = self._press
        
        # Update coordinates
        rline.set_xdata ([event.xdata, event.xdata])
        
        # Update arrow's ends
        headsize = 4
        arrow_lend.set_xy(_triangle_verts_from_xy(max(event.x, x_press)-headsize,
                                                  self._center[1],
                                                  size=headsize-1,
                                                  orientation=0)
                                                  )
        
        arrow_rend.set_xy(_triangle_verts_from_xy(min(event.x, x_press)+headsize,
                                                  self._center[1],
                                                  size=headsize-1,
                                                  orientation=180)
                                                  )
        
        # Update line
        arrow_body.set_xdata([event.x, x_press]) # data are actually pixel due to transform
        
        # Update text position
        x_press_data, y_press_data = self.axes.transData.inverted().transform(self._press)
        text.set_text("%.2f Hz" % abs(event.xdata - x_press_data) )
        text.set_x((event.x + x_press - text.get_window_extent(renderer=self.get_renderer()).width)/2.0)
        
        # Check if visibility is required
        if abs(event.x - x_press) < 10:
            arrow_lend.set_visible(False)
            arrow_rend.set_visible(False)
        else:
            arrow_lend.set_visible(True)
            arrow_rend.set_visible(True)
            
        # Update canvas
        self._update_animated_artists(self._active)
        
    
    def mpl_measure_release_cb(self, event):
        if self.working:
            return
        # Reset active object
        for artist in self._active:
            artist.remove()
        self._active = None
        
        # Disconnect picker and restore default callbacks
        self._restore_default_cb()
        
        # Unset mode
        self.reset_mode()

#===============================================================================
# NMR canvas toolbar
#===============================================================================
class NMRCanvasToolbar(gtk.Toolbar):
    # list of toolitems to add to the toolbar, format is:
    # text, tooltip_text, image, callback(str), type
    # type should be one of button, toggle or None (if separator)
    toolitems = [
                 ("Pan", "Pan", "pan.png", "pan_cb", "toggle"),
                 (None, None, None, None, None), # Separator
                 ("Best zoom", "Zoom on data", "zoom-fit.png", "zoom_fit_cb", "button"),
                 ("Rectangle zoom", "Zoom on rectangle", "zoom-rect.png", "zoom_rect_cb", "toggle"),
                 ("X zoom", "Zoom on X axis", "zoom-x.png", "zoom_x_cb", "toggle"),
                 (None, None, None, None, None), # Separator
                 ("Y", "Change axis used for zoom", "", "_set_axis_zoom", "button"),
                 ("Zoom", "Zoom in", "zoom-in.png", "zoom_in_cb", "button"),
                 ("Zoom", "Zoom out", "zoom-out.png", "zoom_out_cb", "button"),
                 (None, None, None, None, None), # Separator
                 ("Limits", "Set limits and pivot for moment calculation", "lim.png", "set_limits_cb", "toggle"),
                 ("Moments", "Calculate moments", "moments.png", "moments_cb", "button"),
                 (None, None, None, None, None), # Separator
                 ("Depake", "Run FFT-Depaking procedure", "depake.png", "depake_cb", "button"),
                 ]
    
    def __init__(self, canvas, application=None, pix_folder=""):
        # Call ancestor
        gtk.Toolbar.__init__(self)
        self.canvas = canvas
        self.application = application
        
        self.pix_folder = pix_folder
        
        # Internal variables to handle modes
        self._mode = None
        self._modes = {}
        self._calculation_limits = None
        
        # Default axis used for zoom
        self._axis_zoom = "Y"
        
        # Display only icons
        self.set_style(gtk.TOOLBAR_ICONS)
        
        # Add buttons
        self._add_toolitems(self.toolitems)
        
        # Add separator
        toolitem = gtk.SeparatorToolItem()
        self.insert(toolitem, -1)
        toolitem.set_draw(False)
        toolitem.set_expand(True)
        
        # Add message_box
        toolitem = gtk.ToolItem()
        self.message = gtk.Label()
        
        toolitem.add(self.message)
        self.insert(toolitem, -1)
        
        # Connect events
        self.canvas.mpl_connect("motion_notify_event", self.mouse_move_cb)
        self.cid = None
    
    def _add_toolitems(self, toolitems):
        for toolitem in toolitems:
            text, tooltip_text, image, callback, item_type = toolitem
            
            if text is None:
                self.insert(gtk.SeparatorToolItem(), -1)
                continue
            
            # Get image file
            image_file = os.path.join(self.pix_folder,
                                      image)
            
            # Load image if possible
            gtk_image = None
            if os.path.isfile(image_file):
                gtk_image = gtk.Image()
                gtk_image.set_from_file(image_file)
                gtk_image.show()
            
            # Create tool button
            if item_type == "toggle":
                # Create toggle tool button
                gtk_tbutton = gtk.ToggleToolButton()
                
                # Set image
                gtk_tbutton.set_icon_widget(gtk_image)
                
                # Add to modes
                self._modes[text] = gtk_tbutton
            else:
                gtk_tbutton = gtk.ToolButton(gtk_image, text)
            
            # Set tooltip text
            gtk_tbutton.set_tooltip_text(tooltip_text)
            
            # Set callback
            gtk_tbutton.connect("clicked", getattr(self, callback))
            
            # Actually add button
            self.insert(gtk_tbutton, -1)
    
    def _set_axis_zoom(self, widget):
        # Get label from GTK widget
        cur_axis = widget.get_label()
        
        if cur_axis == "Y":
            cur_axis = "X"
        elif cur_axis == "X":
            cur_axis = "XY"
        else:
            cur_axis = "Y"
        
        # Update widget
        widget.set_label(cur_axis)
        
        # Update current axis used for zoom
        self._axis_zoom = cur_axis
        
    def mouse_move_cb(self, event):
        if event.inaxes and event.inaxes.get_navigate():
            try: s = event.inaxes.format_coord(event.xdata, event.ydata)
            except ValueError: pass
            except OverflowError: pass
            else:
                self.message.set_text("%s " % s)
        
    def zoom_fit_cb(self, widget):
        self.canvas.zoom_on_data()
    
    def zoom_rect_cb(self, widget):
        if widget.get_active():
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self.cid = self.canvas.mpl_connect("button_release_event", self.zoom_rect_done_cb)
            
            self._mode = "Rectangle zoom"
            self.canvas.set_mode("box-zoom")
        else:
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self._mode = None
            self.canvas.reset_mode()
        
    def zoom_rect_done_cb(self, event):
        self._modes[self._mode].set_active(False)
        self.canvas.set_mode("box-zoom_confirm")
    
    def zoom_x_cb(self, widget):
        if widget.get_active():
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self.cid = self.canvas.mpl_connect("button_release_event", self.zoom_x_done_cb)
            
            self._mode = "X zoom"
            self.canvas.set_mode("x-box-zoom")
        else:
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self._mode = None
            self.canvas.reset_mode()
        
    def zoom_x_done_cb(self, widget):
        self._modes[self._mode].set_active(False)
        self.canvas.set_mode("x-box-zoom_confirm")
        
    def zoom_in_cb(self, widget):    
        self.canvas.zoom(axis=self._axis_zoom, direction="in")

    def zoom_out_cb(self, widget):    
        self.canvas.zoom(axis=self._axis_zoom, direction="out")
    
    def pan_cb(self, widget):
        if widget.get_active():
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self.cid = self.canvas.mpl_connect("button_release_event", self.pan_done_cb)
            
            self._mode = "Pan"
            self.canvas.set_mode("pan")
        else:
            if self.cid is not None:
                self.canvas.mpl_disconnect(self.cid)
                self.cid = None
            self._mode = None
            self.canvas.reset_mode()
        
    def pan_done_cb(self, widget):
        self._modes[self._mode].set_active(False)
        
    
    def set_limits_cb(self, widget):
        if widget.get_active():
            self._calculation_limits = self.canvas.add_limits()
        else:
            if self._calculation_limits is not None:
                for line in self._calculation_limits:
                    line.remove()
                
                self._calculation_limits = None
                
                self.canvas.draw_idle()
                
    def depake_cb(self, event):
            
        if self.application is not None:            
            # Update cursor
            self.work_start()
            
            # Launch depaking routine
            self.application.depake()
            
            # Update cursor
            self.work_done()
        
    def moments_cb(self, event):
        
        if self._calculation_limits is not None:
            
            llim, rlim, pivot =  self._calculation_limits
            
            llim = llim.get_xdata()[0]
            rlim = rlim.get_xdata()[0]
            pivot = pivot.get_xdata()[0]
            
            if self.application is not None:
                # Update cursor
                self.work_start()
                
                # Launch moment calculation
                self.application.moment_calculation(llim, rlim, pivot)
                
                # Update cursor
                self.work_done()
        else:
            qad_depaker.gtkutil.message_die("Please set calculation limits/pivot first!",
                                            "info")
    
    def work_start(self):
        # Update canvas' cursor
        self.canvas.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
        self.canvas.draw()
        self.canvas.working = True
        
    def work_done(self):
        # Update canvas' cursor
        self.canvas.working = False
        self.canvas.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
        self.canvas.draw_idle()
        
        # Update window's cursor
        self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))
        
        
